@@ -1,26 +1,27 @@
 #!/usr/bin/perl
 
 # apt-cacher-cleanup.pl
-# Script to clean the cache for the Apt-cacher package caching system.
 #
+# Script to clean the apt-cacher cache.
+#
+# Copyright (C) 2007-11, Mark Hindley <mark@hindley.org.uk>
 # Copyright (C) 2005, Eduard Bloch <blade@debian.org>
 # Copyright (C) 2002-03, Jonathan Oxer <jon@debian.org>
 # Portions  (C) 2002, Jacob Lundberg <jacob@chaos2.org>
+#
 # Distributed under the terms of the GNU Public Licence (GPL).
 
-
-# do locking, not losing files because someone redownloaded the index files
-# right then
-
 use strict;
-use lib '/usr/share/apt-cacher/';
+use warnings;
+use lib '/usr/share/apt-cacher';
 
+use sigtrap qw(die normal-signals);
 use Cwd;
 use Fcntl qw/:DEFAULT :flock F_SETFD/;
-use Getopt::Long qw(:config no_ignore_case bundling pass_through);
-use Digest::SHA1;
+use Getopt::Long qw(:config no_ignore_case);
+use Digest::SHA;
 use HTTP::Date;
-use FreezeThaw qw(freeze thaw);
+use HTTP::Response;
 use IO::Uncompress::AnyUncompress qw($AnyUncompressError);
 use IO::Compress::Bzip2;
 use IO::Compress::Gzip;
@@ -42,41 +43,60 @@ my %options = (
     "n|nice" => \$nice_mode,
     "v|verbose" => \$verbose,
     "f|force" => \$force,
-    "c|config-file=s" => \$configfile,
+    "c|cfg|conf=s" => \$configfile,
     "s|simulate" => \$sim_mode,
     "o|offline" => \$offline,
     "p|pdiff" => \$pdiff_mode,
     "r|recover" => \$db_recover,
-    "d|db:s" => \@db_mode
+    "d|db=s{,}" => \@db_mode
 );
 
-my @savedARGV = @ARGV; # Save a copy in case required for rexec
-&help unless ( GetOptions(%options));
+{ # Scoping block
+    local @ARGV = @ARGV; # Use a local copy in case required for rexec
+    if (!GetOptions(%options) || $help) {
+	die <<EOM
+Usage: $0  [-c|cfg|conf <configfile>] [-n|nice] [-s|simulate]
+ [-v|verbose] [-o|offline] [-f|force] [-r|recover] ][-d|db <command> [<arg>]]
+
+Options:
+ -n	Renice to lowest priority and continue.
+ -s	Simulate and just show what would be done.
+ -o	Don't update index files. Overrides offline_mode from configfile.
+ -p	Try to update index files by patching.
+ -v	Verbose.
+ -f	Force execution, disable sanity checks.
+ -r	Attempt recovery of corrupt checksum database.
+
+ -d <command> [<arg>]	Manipulate checksum database.
+    Available commands are:
+	dump		print db contents
+	search <arg>	print entries matching regexp
+	delete <arg>	delete entries matching regexp
+	import		read new checksum data from Packages and Sources files in cache_dir
+	compact		clean and compact database
+        verify		verify the database
+EOM
+    }
+}
 
 if ($sim_mode) {
   $verbose = 1;
   print "Simulation mode. Just printing what would be done.\n";
 }
-&help if ($help);
 
 #############################################################################
 ### configuration ###########################################################
 # Include the library for the config file parser
-require 'apt-cacher-lib.pl';
+require('apt-cacher-lib.pl');
 # Read in the config file and set the necessary variables
 
 # $cfg needs to be global for setup_ownership
-our $cfg;
-
-eval {
-	 $cfg = read_config($configfile);
-};
+our $cfg = eval{ read_config($configfile) };
 
 # not sure what to do if we can't read the config file...
-die "Could not read config file: $@" if $@;
+die "Could not read configuration file '$configfile': $@" if $@;
 
-my $globlockfile="$cfg->{cache_dir}/private/exlock";
-define_global_lockfile($globlockfile);
+private_config();
 
 # check whether we're actually meant to clean the cache
 if ( $cfg->{clean_cache} ne 1 ) {
@@ -85,73 +105,52 @@ if ( $cfg->{clean_cache} ne 1 ) {
 }
 
 # change uid and gid if root and another user/group configured
-if ($cfg->{user} && $cfg->{user} !~ 'root' && !$> 
-    or $cfg->{group} && $cfg->{group} !~ 'root' && !$) =~ /^0/){
+if (($cfg->{user} && $cfg->{user} !~ 'root' && !$> )
+    || ($cfg->{group} && $cfg->{group} !~ 'root' && !$) =~ /^0/)){
     printmsg("Invoked as root, changing to $cfg->{user}:$cfg->{group} and re-execing.\n");
     setup_ownership($cfg);
     # Rexec to ensure /proc/self/fd ownerships correct which are needed for red
     # patching with pdiffs
-    exec($0, @savedARGV) or die "Unable to rexec $0: $!\n";
+    exec($0, @ARGV) or die "Unable to rexec $0: $!\n";
 }
 # Output data as soon as we print it
-$| = 1;
+local $| = 1;
 
 setpriority 0, 0, 20 if $nice_mode;
 
-sub help {
-    die <<EOM
-    Usage: $0 [ -n ] [ -s|v ] [ -o ] [ -f ] [-d command [arg]] [ -c configfile ]
-    -n : nice mode, renice to lowest priority and continue
-    -s : simulate mode, just print what would be done to package files
-    -o : offline mode, don't update index files. Overrides offline_mode from configfile
-    -p : patch mode [experimental]. Try to update index files by patching
-    -v : verbose mode
-    -r : recover, attempt recovery of corrupt checksum database
-    -d command [arg]: db mode -- manipulate checksum database.
-       Available commands are:
-        dump:	      print db contents
-	search [arg]: print entries matching regexp
-	delete [arg]: delete entries matching regexp
-	import:	      read new checksum data from Packages and Sources files in cache_dir
-	compact:      clean and compact database
-        verify:	      verify the database
-    -f : force executing, disable sanity checks
-EOM
-    ;
-}
-
 sub printmsg {
-   print @_ if $verbose;
+    my @args = @_;
+    my $ret;
+    $ret = print @args if $verbose;
+    return $ret;
 }
 
 sub get {
-    my ($file, $use_complete) = @_;
+    my ($file, $use_url) = @_;
     my $path_info;
-    my $fh;
-    if($use_complete && -s "../private/$file.complete") {
-    # use path is stored in complete file
-	open(my $tmp, "../private/$file.complete");
-	$path_info=<$tmp>;
-	close $tmp;
+    # use path if stored in cached header/complete file
+    if($use_url && (my $url = get_original_url($file))) {
+	$path_info= $url;
     }
     else {
 	$path_info=$file;
 	$path_info=~s/^/\//;
 	$path_info=~s/_/\//g;
     }
-    open($fh, "| REMOTE_ADDR=CLEANUPREFRESH /usr/share/apt-cacher/apt-cacher -i -c $configfile >/dev/null");
+    open(my $fh, '|-', "REMOTE_ADDR=CLEANUPREFRESH /usr/share/apt-cacher/apt-cacher -i -c $configfile >/dev/null") || die "Failed to open refresh pipe: $!";
     printmsg "GET $path_info\n";
     print $fh "GET $path_info\r\nCache-Control: max-age=0\r\nConnection: Close\r\n\r\n";
     close($fh);
     if($? && ! $force) {
 	die "Unable to update $path_info.\nCleanup aborted to prevent deletion of cached data.\n";
     }
+    return $?;
 }
 
 sub pdiff {
-    my $name = $_[0];
+    my ($name) = @_;
     if (!-f $name) {
-	warn ("$name not found\n");
+	warn ("File $name not found\n");
 	return;
     }
 
@@ -165,7 +164,7 @@ sub pdiff {
 
     for ($release, $diffindex) {
 	if (!-f $_ && !$offline) { # Refresh unless offline
-	    &get($_);
+	    get($_);
 	}
 	if (!-f $_) {
 	    printmsg("$_ not available, aborting patch\n");
@@ -180,11 +179,11 @@ sub pdiff {
 	s/_/\//g;
     }
 #    printmsg "Searching $release for $diffindex_patt and $name_patt\n";
-    open(my $rfh, $release) || die "Unable to open $release: $!";
+    open(my $rfh, '<', $release) || die "Unable to open $release: $!";
     flock($rfh, LOCK_SH);
 
     my ($diffindex_sha1, $name_sha1, $name_size);
-    while(<$rfh>) {
+    while (<$rfh>) {
 	if (/^\s(\w{40})\s+\d+\s$diffindex_patt\n/) {
 	    $diffindex_sha1 = $1;
 #	    printmsg "Found! $diffindex_patt $1\n";
@@ -203,7 +202,7 @@ sub pdiff {
 	return;
     }
 
-    my $sha1 = Digest::SHA1->new;
+    my $sha1 = Digest::SHA->new(1); # SHA1
     my $digest;
 
     # Check size first
@@ -211,9 +210,9 @@ sub pdiff {
 	printmsg ("$name matches size in $release, going on to check SHA1..\n");
 	
 	# Check SHA1 only if size correct
-	open(my $nfh, $name)|| die "Unable to open file $name for locking: $!";
+	open(my $nfh, '<', $name)|| die "Unable to open file $name for locking: $!";
 	flock($nfh, LOCK_SH);	
-	$digest = $sha1->addfile(*$nfh)->hexdigest;
+	$digest = $sha1->addfile($nfh)->hexdigest;
 	flock($nfh, LOCK_UN);
 	close($nfh,);
 	if ($digest eq $name_sha1) {
@@ -228,10 +227,10 @@ sub pdiff {
 	printmsg ("$name size not latest, proceeding with patch\n");
     }
 
-    open(my $lck, $name)|| die "Unable to open file $name for locking: $!";
+    open(my $lck, '<', $name)|| die "Unable to open file $name for locking: $!";
     flock($lck, LOCK_SH);
 
-    my $raw = new IO::Uncompress::AnyUncompress $name
+    my $raw = IO::Uncompress::AnyUncompress->new($name)
       or die "Decompression failed: $AnyUncompressError\n";
 
     open (my $tfh, "+>", undef)|| die "Unable to open temp file: $!";
@@ -253,15 +252,15 @@ sub pdiff {
     $digest = $sha1->hexdigest;
 #    printmsg "$basename SHA1: $digest\n";
 
-    # Read diff.Index
+    # Read diff_Index
     my (@hist, @patch);
 
-    open(DIFFIN, $diffindex) || die("Cannot open $diffindex: $!");
-    flock(DIFFIN, LOCK_EX);
-    my $diffindex_digest = $sha1->addfile(*DIFFIN)->hexdigest;
+    open(my $diffin_fh, '<', $diffindex) || die("Cannot open $diffindex: $!");
+    flock($diffin_fh, LOCK_EX);
+    my $diffindex_digest = $sha1->addfile($diffin_fh)->hexdigest;
     if ($diffindex_digest ne $diffindex_sha1) {
-	flock(DIFFIN, LOCK_UN);
-	close (DIFFIN);
+	flock($diffin_fh, LOCK_UN);
+	close ($diffin_fh);
 	if ($force) {
 	    warn "$diffindex incorrect SHA1: expected $diffindex_sha1, got $diffindex_digest. Continuing anyway as --force specified\n";
 	}
@@ -270,31 +269,31 @@ sub pdiff {
 	return;
 	}
     }
-    seek(DIFFIN,0,0); # rewind
-    my $curr= <DIFFIN>; # read first line
+    seek($diffin_fh,0,0); # rewind
+    my $curr= <$diffin_fh>; # read first line
     chomp $curr; # remove trailing \n
 #    printmsg "$diffindex: $curr\n";
     my ($target_sha1, $target_size) = (split (/\s+/,$curr))[1,2];
     if ($digest eq $target_sha1) { # check this matches /SHA1/
 	printmsg "SHA1 match: $name already up to date\n";
-	flock(DIFFIN, LOCK_UN);
-	close (DIFFIN);
+	flock($diffin_fh, LOCK_UN);
+	close ($diffin_fh);
 	return 1; # success
     }
     else {
-	while (<DIFFIN>) {
+	while (<$diffin_fh>) {
 	    next if (/^SHA1-History:/); # skip header
 	    last if (/^SHA1-Patches:/);# end of history
 	    push @hist, $_;
 	    next;
 	}
-	while (<DIFFIN>) {
+	while (<$diffin_fh>) {
 	    push @patch, $_; # To EOF
 	    next;
 	}
     }
-    flock(DIFFIN, LOCK_UN);
-    close (DIFFIN);
+    flock($diffin_fh, LOCK_UN);
+    close ($diffin_fh);
 
     my $diff;
     my $count=0;
@@ -316,13 +315,13 @@ sub pdiff {
 
     my $diffs=''; # Initialise to work around perl bug giving "Use of uninitialized value error"
 	
-    open(DIFFS, ">", \$diffs);
+    open(my $diffs_fh, ">", \$diffs) || die "Failed to open in memory diff file: $!";
     for (@patch[$diff .. $#patch]) {
 	my ($pdiffsha1, $size, $suff) = split;
-	my $pdiff = $basename.".diff_".(split)[2].'.gz';
+	my $pdiff = "$basename.diff_$suff.gz";
 	if (!-f $pdiff) {
 	    if (!$offline) {
-		&get($pdiff);
+		get($pdiff);
 	    }
 	    if (!-f $pdiff) {
 		warn("$pdiff not available, aborting patch");
@@ -330,11 +329,11 @@ sub pdiff {
 	    }
 	}
 	printmsg "Reading $pdiff\n";
-	my $pdfh = new IO::Uncompress::AnyUncompress $pdiff
+	my $pdfh = IO::Uncompress::AnyUncompress->new($pdiff)
 	  or die "Decompression failed: $AnyUncompressError\n";
 	while (<$pdfh>) {
 	    last if $AnyUncompressError;
-	    print DIFFS $_;
+	    print $diffs_fh $_;
 	    $sha1->add($_);
 	}
 	close($pdfh);
@@ -350,16 +349,16 @@ sub pdiff {
 	    return;
 	}
     }
-    print DIFFS "w\n"; # append ed write command
-    close(DIFFS);
+    close($diffs_fh);
 
     fcntl($tfh, F_SETFD, 0)
       or die "Can't clear close-on-exec flag on temp filehandle: $!\n";
     my $cwd = Cwd::cwd(); # Save
     chdir '/dev/fd' or  die "Unable to change working directory: $!";
-    open(my $patchpipe, "| $patchprog ".fileno($tfh)) ||  die "Unable to open pipe for patch: $!";
+    open(my $patchpipe, '|-', "$patchprog ".fileno($tfh)) ||  die "Unable to open pipe for patch: $!";
     printmsg "Patching $name with $patchprog\n";
     print $patchpipe $diffs;
+    print $patchpipe "w\n"; # ed write command
     close($patchpipe);
     chdir $cwd or die "Unable to restore working directory: $!"; # Restore
     my $rstat =($? >> 8);
@@ -373,7 +372,7 @@ sub pdiff {
 	return;
     }
     seek($tfh,0,0); # rewind
-    $sha1->addfile(*$tfh);
+    $sha1->addfile($tfh);
     $digest=$sha1->hexdigest;
     if ($digest eq $target_sha1) {
 	printmsg "Success! SHA1: $digest\n";
@@ -383,54 +382,41 @@ sub pdiff {
 	else {
 	    my $destfile = "../temp/".$basename.".new".$type;
 	    printmsg "Saving as $destfile\n";
-	    if (-f $destfile) {
-		printmsg "Warning: $destfile already exists\n";
-	    }
+	    printmsg "Warning: $destfile already exists\n" if -f $destfile;
 	    seek($tfh,0,0); # rewind
 	    my ($z,$encoding) = ($name=~/bz2$/ ?
-				 ((new IO::Compress::Bzip2 $destfile), "x-bzip2") :
+				 ((IO::Compress::Bzip2->new($destfile)), "x-bzip2") :
 				 ($name=~/gz$/ ?
-				  ((new IO::Compress::Gzip $destfile, -Level => 9), "x-gzip") :
-				  new IO::Handle $destfile));
+				  ((IO::Compress::Gzip->new($destfile, -Level => 9)), "x-gzip") :
+				  IO::Handle->new($destfile)));
 
-	    while(<$tfh>) {
+	    while (<$tfh>) {
 		$z->print($_);
 	    }
-	    close($z);
-	    my @info = stat($destfile);
+	    $z->flush();
+	    my $size = -s $destfile;
 	    my $datestring = HTTP::Date::time2str;
-	    open(my $hfh, ">", "$destfile.header") || die "Unable to open header file: $!\n";
-	    printmsg "Writing header\n";
-	    print $hfh <<EOF;
-HTTP/1.0 200 OK
-Connection: Keep-Alive
-Accept-Ranges: bytes
-Content-Length: $info[7]
-Content-Type: text/plain
-Last-Modified: $datestring
-EOF
-	    if ($encoding) {print $hfh "Content-Encoding: $encoding\n"};
-	    close($hfh);
-	    # Get global lock and both locks for files
-	    set_global_lock(": copy patched file");
-	    open(my $pfh, $name)|| die "Unable to open packages/$name for locking: $!";
-	    flock($pfh, LOCK_EX);
-	    open($hfh, "../headers/$name")|| die "Unable to open headers/$name for locking: $!";
-	    flock($hfh, LOCK_EX);
+	    my $response = HTTP::Response->new(200, 'OK', ['Date' => $datestring,
+							   'Content-Length' => $size,
+							   'Content-Type' => 'text/plain',
+							   'Last-Modified' => $datestring]);
+	    $response->header('Content-Encoding' => $encoding) if $encoding;;
+	    write_header("$destfile.header", $response);
+	    # Get global lock and copy
+	    set_global_lock('copy patched file');
 	    printmsg("Linking to $name\n");
 	    unlink $name,  "../headers/$name";
 	    link $destfile, $name or die "Link $destfile failed: $!";
 	    link "$destfile.header", "../headers/$name" || die "Link $destfile.header failed: $!";
 	    printmsg "Unlink temporary files\n";
 	    unlink $destfile, "$destfile.header";
-	    flock($hfh, LOCK_UN);
-	    flock($pfh, LOCK_UN);
 	    release_global_lock();
 	    # Read checksums
 	    if ($cfg->{checksum}) {
 		printmsg ("Importing new checksums from patched $name\n");
-		import_sums($tfh,&db);
+		import_sums($z);
 	    }
+	    close($z);
 	}
     }
     else {
@@ -444,24 +430,33 @@ EOF
 # Calls _db_compact to do the work and reports results
 # Arg: DB handle ref
 sub db_compact {
+    my ($dbh) = @_;
     printmsg "Compacting checksum database....\n";
-    my ($status, $results) = &_db_compact($_[0]);
-    if ($$status) {
-	printmsg "db_compact failed: $$status\n";
-    }
-    else {
-	printmsg " Compacted ". $results->{compact_pages_free} ." pages\n Freed ". $results->{compact_pages_truncated} ." pages\nDone!\n";
+    while (my ($status, $results) = _db_compact($dbh)) {
+	if ($$status) {
+	    printmsg "db_compact failed: $$status\n";
+	    last;
+	}
+	else {
+	    printmsg " Compacted ". $results->{compact_pages_free} ." pages\n Freed ". $results->{compact_pages_truncated} ." pages\n";
+	    if ($results->{compact_pages_free} + $results->{compact_pages_truncated} == 0) {
+		printmsg "Done!\n";
+		last;
+	    }
+	}
     }
+    return;
 }
 
 #############################################################################
 # Manipulate checksum database
 if (@db_mode || $db_recover){
 
-    my $ok_chars = '-a-zA-Z0-9+_.~'; # Acceptable characters for user input
+    my $ok_chars = '-a-zA-Z0-9+_.,~^$*?{}[]()'; # Acceptable characters for user input
+    $ok_chars .= '/' if $cfg->{distinct_namespaces};
     print "Checksum database mode\n";
 
-    require 'apt-cacher-lib-cs.pl';
+    load_checksum();
 
     if (!$cfg->{checksum}) {
 	die "$0: checksumming not enabled. Use --force to override\n" if !$force;
@@ -471,74 +466,72 @@ if (@db_mode || $db_recover){
 
     if ($db_recover) {
 	printmsg "Running database recovery...";
-	&db_recover;
+	db_recover();
 	printmsg "Done!\n";
     }
 
-    chdir "$cfg->{cache_dir}/packages" || die "Unable to enter cache dir: $!";
-
-    my $dbh = ${&db}|| die "Unable to init db: $!\n";
+    chdir "$cfg->{cache_dir}/packages" || die "Unable to enter cache package dir: $!";
 
   SWITCH:
-    for (@db_mode) {
-	/^import/ && do {
-	    for (<*es.bz2>, <*es.gz>, <*es>, <*Release>, <*diff_Index>) {
+    while (local $_ = shift @db_mode) {
+	/^import$/ && do {
+	    foreach (glob('*es.bz2 *es.gz *es *Release *diff_Index')) {
 		printmsg "Importing checksums from $_\n";
-		&import_sums($_,\$dbh) if !$sim_mode;
+		import_sums($_) if !$sim_mode;
 	    }
 	    next SWITCH;
 	};
-	/^compact/ && do {
-	    &db_compact(\$dbh);
+	/^compact$/ && do {
+	    db_compact(db());
 	    next SWITCH;
 	};
-	/^(?:dump|search)/ && do {
+	/^(?:dump|search)$/ && do {
 	    my $re;
-	    if (/^search/){
-		$re=shift;
+	    if (/^search$/){
+		$re = shift @db_mode;
 		die "No search expression given\n" if !$re;
 		die "Invalid character '$1' in search\n" if $re =~ /([^$ok_chars])/o; # sanitize
 	    }
-	    my $cursor = get_cursor($dbh);
-	    my ($filename,$data)=('','');
-	    while (&cursor_next($cursor, \$filename, \$data) == 0)
+	    my $cursor = get_cursor(db());
+	    my ($filename,$data) = ('','');
+	    while (cursor_next($cursor, \$filename, \$data) == 0)
 	      {
 		  next if /^search/ && $filename !~ /$re/;
 		  print "$filename\n";
 		  my $href = hashify(\$data);
 		  while (my ($k,$v) = each %$href) {
-		      $v='' if ! defined $v;
+		      $v = '' if ! defined $v;
 		      print " $k: $v\n";
 		  }
 	      }
 	    next SWITCH;
 	};
-	/^delete/ && do {
-	    my $re=shift;
+	/^delete$/ && do {
+	    my $re = shift @db_mode;
 	    die "No give regex to match files to delete\n" if !$re;
 	    die "Invalid character '$1' in pattern\n" if $re =~ /([^$ok_chars])/o; # sanitize
-	    my $cursor = get_cursor($dbh,1);
-	    my ($filename,$data)=('','');
-	    while (&cursor_next($cursor, \$filename, \$data) == 0)
+	    my $cursor = get_cursor(db(),1);
+	    my ($filename,$data) = ('','');
+	    while (cursor_next($cursor, \$filename, \$data) == 0)
 	      {
 		next if $filename !~ /$re/;
 		printmsg "Deleting data for $filename\n";
-		$cursor->c_del and warn "c_del failed: $!" if !$sim_mode;
+		$cursor->c_del && warn "c_del failed: $BerkeleyDB::Error" if !$sim_mode;
 	    }
 	    next SWITCH;
 	};
-	/^verify/ && do {
+	/^verify$/ && do {
 	    printmsg "Waiting for exclusive lock...";
-	    if (&db_flock(LOCK_EX)){
+	    if (db_flock(LOCK_EX)){
 		printmsg "Got it!\nVerifying database...";
-		printmsg &db_verify("$cfg->{cache_dir}/sums.db", &temp_env) ? "Failed! $!\n" : "Passed!\n";
+		printmsg db_verify("$cfg->{cache_dir}/sums.db", temp_env()) ? "Failed! $!\n" : "Passed!\n";
 	    }
 	    else {
 		warn "Unable to get exclusive database lock: $!\n";
 	    }
 	    next SWITCH;
 	};
-	warn "Unknown command $_ ".(shift)."\n";
+	warn "Unknown command $_ \n";
 	next SWITCH;
     }
 
@@ -562,45 +555,54 @@ if (defined $cfg->{offline_mode} && $cfg->{offline_mode}) {
 	$offline = 1;
 }
 
-use DB_File;
-tie my %valid, 'DB_File';
+use GDBM_File;
+open(my $tmpfile, "+>", undef) or die $!;
+tie my %valid, 'GDBM_File', '/dev/fd/'.fileno($tmpfile), &GDBM_NEWDB|&GDBM_FAST, oct(600); # Does a separate open
+close($tmpfile); # So we can close this
 
 my $tempdir="$cfg->{cache_dir}/temp";
 mkdir $tempdir if !-d $tempdir;
 die "Could not create tempdir $tempdir\n" if !-d $tempdir;
-unlink (<$tempdir/*>);
+unlink (glob("$tempdir/*"));
 
 ### Preparation of the package lists ########################################
 
-chdir "$cfg->{cache_dir}/packages" && -w "." || die "Could not enter the cache dir";
+chdir "$cfg->{cache_dir}/packages" && -w "." || die "Could not enter the cache dir: $!";
 
 if($> == 0 && !$cfg->{user} && !$force) {
-    die "Running $0 as root\nand no effective user has been specified. Aborting.\nPlease set the effective user in $configfile\n";
+    die "Running $0 as root\nand no effective user has been specified. Aborting.\nPlease set the effective user in $configfile or use --force to ignore\n";
 }
 
-# Ensure corresponding Packages/Sources is present for each diff_Index
-DIFFINDEX:
-for (<*diff_Index>) {
-    if (/^(.+)\.diff_Index/) {
-	printmsg "Checking for $1 for $_\n";
-	for ($1,"$1.gz", "$1.bz2") {
+# Try to ensure corresponding Packages/Sources is present for each diff_Index
+# and Release for each Packages/Sources
+{
+    my %missing;
+  CHECKFILE:
+    foreach (glob('*diff_Index *{Packages,Sources}{,.gz,.bz2}')) {
+	my $file = $_;
+	if (s/\.diff_Index$/{,.gz,.bz2}/) {
+	    printmsg "Checking for $_ for $file\n";
+	}
+	elsif (s/(?:dists_[^_]+_(?:updates_)?\K(?:[^_]+_){2})?(?:Packages|Sources)(?:\.(?:bz2|gz))?$/{In,}Release/) {
+	    printmsg "checking for $_ for $file\n";
+	}
+	foreach (glob) {
 	    if ( -f $_ ) {
-		printmsg "Found $_\n";
-		next DIFFINDEX;
+		print "Found $_\n";
+		next CHECKFILE;
 	    }
 	}
-	printmsg ("Not found. Downloading\n");
-	# Might as well use bzipped files
-	&get("$1.bz2");
+	$missing{(glob)[-1]} = 1;
     }
+    get($_) foreach keys %missing;
 }
 
 # file state decisions, lock that area
-set_global_lock(": file state decision");
-my @ifiles=(<*Release>, <*_Index> , <*es.gz>, <*es.bz2>, <*es>);
+set_global_lock('file state decision');
+my @ifiles = glob('*{Release,diff_Index} *{Packages,Sources}{,.gz,.bz2}');
 release_global_lock();
 
-for my $file (@ifiles) {
+foreach my $file (@ifiles) {
 
    # preserve the index files
    $valid{$file}=1;
@@ -614,7 +616,7 @@ for my $file (@ifiles) {
    # If patching failed download them, unless offline
    if (!$patched) {
        if(!$offline) {
-	   &get($file);
+	   get($file);
        }
        else {
 	   printmsg "Offline: Reusing existing $file\n";
@@ -622,13 +624,13 @@ for my $file (@ifiles) {
    }
 }
 
-for my $file (@ifiles) {
+foreach my $file (@ifiles) {
     printmsg "Reading: $file\n";
 
     # get both locks and create a temp. copy
     my $tmpfile= "$tempdir/$file";
-    set_global_lock(": temporary copy");
-    open(my $lck, $file) || do {
+    set_global_lock('temporary copy');
+    open(my $lck, '<', $file) || do {
 	release_global_lock();
 	print ("Error: cannot open $file for locking: $!\n");
 	next;
@@ -659,97 +661,115 @@ for my $file (@ifiles) {
 printmsg "Found ".scalar (keys %valid)." valid file entries\n";
 #print join("\n",keys %valid);
 
+# Build a source package version reverse hash for changelog validation from the .dsc files
+printmsg "Building source package file/version table\n";
+my %svrhash = map {m#([-+.a-z0-9]+_(?:\d:)?[-+.~a-zA-Z0-9]+)\.dsc# && $1 => 1} keys %valid;
+
 # Remove old checksum data
+load_checksum(); # Will disable checksum if BerkeleyDB not available
+
 if ($cfg->{checksum}) {
-   require 'apt-cacher-lib-cs.pl';
 
-   my $dbh = ${&db};
+    my $dbh = db();
 
-   my $do_compact;
-   $dbh && do {
-       printmsg "Removing expired entries from checksum database\n";
-
-
-       my $cursor = get_cursor($dbh,1);
-       my ($filename,$data)=('','');
-       while (&cursor_next($cursor, \$filename, \$data) == 0)
-	 {
-	   next if defined $valid{$filename};
-	   printmsg "Deleting checksum data for $filename\n";
-	   $cursor->c_del and warn "c_del failed: $!" if !$sim_mode;
-	   $do_compact = 1;
-       }
-       &db_compact(\$dbh) if $do_compact || $pdiff_mode;
-   };
+    my $do_compact;
+    $dbh && do {
+	printmsg "Removing expired entries from checksum database\n";
+	
+	my $cursor = get_cursor($dbh,1);
+	my ($filename,$data)=('','');
+	while (cursor_next($cursor, \$filename, \$data) == 0)
+	  {
+	      next if defined $valid{$filename};
+	      printmsg "Deleting checksum data for $filename\n";
+	      $cursor->c_del && warn "c_del failed: $BerkeleyDB::Error" if !$sim_mode;
+	      $do_compact = 1;
+	  }
+	db_compact($dbh) if $do_compact || $pdiff_mode;
+    };
 }
 
-for(<*.deb>, <*.udeb>, <*.bz2>, <*.gz>, <*.dsc>) {
-    # should affect source packages but not index files which are added to the
-    # valid list above
+# Clean package directory
+foreach (glob('*{,/*}')) {
+    next if -d; # Skip directories
+    if (/([-+.a-z0-9]+_(?:\d:)?[-+.~a-zA-Z0-9]+)_changelog$/ && !$svrhash{$1}) {
+	unlink $_, "../headers/$_" unless $sim_mode;
+	printmsg "Removing expired changelog: $_ and company...\n";
+	next;
+    }
+    else {
+	next unless is_package_file($_); # Package files only	
+    }
     if(! defined($valid{$_})) {
 	unlink $_, "../headers/$_", "../private/$_.complete" unless $sim_mode;
 	printmsg "Removing file: $_ and company...\n";
     }
     else {
-	# Verify SHA1 checksum for uncompressed files only
-	if(/\.u?deb$/) {
-	    my $target_sum = hashify(\$valid{$_})->{sha1};
-	    next unless $target_sum;
-#	    print "Validating SHA1 $target_sum for $_\n";
-	    open(my $fh, $_) || die "Unable to open file $_ to verify checksum: $!";
-	    flock($fh, LOCK_EX);
-	    if (Digest::SHA1->new->addfile(*$fh)->hexdigest ne $target_sum) {
-		unlink $_, "../headers/$_", "../private/$_.complete" unless $sim_mode;
-		printmsg "Checksum mismatch: $_, removing\n";
-	    }
-	    flock($fh, LOCK_UN);
-	    close $fh;
+	# Verify SHA1 checksum
+	my $target_sum = hashify(\$valid{$_})->{sha1};
+	next unless $target_sum;
+	# print "Validating SHA1 $target_sum for $_\n";
+	open(my $fh, '<', $_) || die "Unable to open file $_ to verify checksum: $!";
+	flock($fh, LOCK_EX);
+	if ((my $sha1 = Digest::SHA->new(1)->addfile($fh)->hexdigest) ne $target_sum) {
+	    unlink $_, "../headers/$_", "../private/$_.complete" unless $sim_mode;
+	    printmsg "Checksum mismatch ($target_sum <=> $sha1): $_, removing\n";
 	}
+	flock($fh, LOCK_UN);
+	close $fh;
     }
 }
 
-# similar thing for possibly remaining cruft
-chdir "$cfg->{cache_dir}/headers" && -w "." || die "Could not enter the cache dir";
+# Clean header directory
+chdir "$cfg->{cache_dir}/headers" && -w "." || die "Could not enter the cache header dir: $!";
 
-# headers for previously expired files
-for(<*.deb>, <*.udeb>, <*.bz2>, <*.gz>, <*.dsc>) {
-   if(! defined($valid{$_})) {
-      unlink $_, "../private/$_.complete" unless $sim_mode;
-      printmsg "Removing expired headers: $_ and company...\n";
-   }
+foreach (glob('*{,/*}')) {
+    next if -d; # Skip directories
+    if((is_package_file($_) && !defined($valid{$_})) # Not indexed
+       || !-e "../packages/$_") { # No corresponding package
+	unlink $_, "../private/$_.complete" unless $sim_mode;
+	printmsg "Removing expired headers: $_ and company...\n";
+	next;
+    }
+    my $resp = read_header($_);
+    next if $resp && $resp->is_success; # Don't cache errors any more
+    printmsg "Removing cached error/invalid response: $_\n";
+    delete $valid{$_};
+    unlink $_ unless $sim_mode;
 }
 
-# also remove void .complete files, created by broken versions of apt-cacher in rare conditions
-chdir "$cfg->{cache_dir}/private" && -w "." || die "Could not enter the cache dir";
-for(<*.deb.complete>, <*.udeb.complete>, <*.bz2.complete>, <*.gz.complete>, <*.dsc.complete>) {
+# Remove .complete files, for which we no longer have cached data.  No new
+# complete files are being created, but they do contain the URL that was used,
+# so keep those for now
+chdir "$cfg->{cache_dir}/private" && -w "." || die "Could not enter the cache private dir: $!";
+foreach (glob('*.complete')) {
    s/.complete$//;
-   if(! (defined($valid{$_}) && -e "../packages/$_" && -e "../headers/$_") ) {
+   if(!(defined($valid{$_}) && -e "../packages/$_" && -e "../headers/$_")) {
       printmsg "Removing: $_.complete\n";
       unlink "$_.complete" unless $sim_mode;
    }
+   elsif (-e "../headers/$_") {
+       my $resp = read_header($_);
+       next if !$resp || !$resp->header('X-AptCacher-URL');
+       # Remove complete files if we now have the upstream URL in the headers
+       printmsg "Removing redundant $_.complete\n";
+       unlink "$_.complete" unless $sim_mode;
+   }
 }
 
 # last step, kill some zombies
-
-my $now = time();
-for(<*.notify>) {
-    my @info = stat($_);
-    # even the largest package should be downloadable in two days or so
-    if(int(($now - $info[9])/3600) > 48) {
-	printmsg "Removing orphaned notify file: $_\n";
-	unlink $_ unless $sim_mode;
-    }
+foreach (glob('*.notify')) {
+    printmsg "Removing obsolete notify file: $_\n";
+    unlink $_ unless $sim_mode;
 }
 
-#&set_global_lock(": cleanup zombies");
+chdir "$cfg->{cache_dir}/packages" || die "Could not enter the cache package dir: $!";
 
-chdir "$cfg->{cache_dir}/packages";
-
-for(<*>) {
+foreach (glob('*{,/*}')) {
     # must be empty and not complete and being downloaded right now
     if(-z $_) {
 	my $fromfile;
-	if(open($fromfile, $_) && flock($fromfile, LOCK_EX|LOCK_NB)) {
+	if(open($fromfile, '<', $_) && flock($fromfile, LOCK_EX|LOCK_NB)) {
 	    # double-check, may have changed while locking
 	    if(-z $_) {
 		printmsg "Removing zombie files: $_ and company...\n";
@@ -761,4 +781,4 @@ for(<*>) {
     }
 }
 
-unlink (<$tempdir/*>);
+unlink (glob("$tempdir/*"));
