@@ -1,4 +1,4 @@
-#!/usr/bin/perl -w
+#!/usr/bin/perl
 
 # apt-cacher-report.pl
 # Script to generate usage reports for the Apt-cacher package caching system.
@@ -6,37 +6,57 @@
 # Copyright (C) 2002,2004 Jonathan Oxer <jon@debian.org>
 # Distributed under the terms of the GNU Public Licence (GPL).
 
-#use strict;
+use strict;
+use warnings;
 #############################################################################
 ### configuration ###########################################################
 # Include the library for the config file parser
-require '/usr/share/apt-cacher/apt-cacher-lib.pl';
+require('/usr/share/apt-cacher/apt-cacher-lib.pl');
 use POSIX qw(strftime);
 
 
 # Read in the config file and set the necessary variables
 my $configfile = '/etc/apt-cacher/apt-cacher.conf';
 
-# Needs to be global for &setup_ownership
-our $cfg;
-eval {
-        $cfg = read_config($configfile);
-};
+while(scalar (@ARGV)) {
+    my $arg=shift(@ARGV);
+    if($arg eq '-c') {
+	$configfile=shift(@ARGV) || die '-c option requires an argument';
+	die "$configfile not a file" if ! -f $configfile;
+	die "$configfile unreadable" if ! -r $configfile;
+    }
+    elsif($arg eq '-h' || $arg eq '--help') {
+	print <<EOM;
+USAGE: $0 <options>
+Options:
+
+-c configfile   Custom config file (default: $configfile)
+-h|--help	Print this help message
+EOM
+	exit(0);
+    }
+    else {
+	die "Unknown parameter $arg\n";
+    }
+}
+
+# Needs to be global for setup_ownership()
+our $cfg = eval{ read_config($configfile) };
 
 # not sure what to do if we can't read the config file...
 die "Could not read config file: $@" if $@;
 
 # check whether we're actually meant to generate a report
-if ( $$cfg{generate_reports} ne 1 ){
+if ( $cfg->{generate_reports} ne 1 ){
 	exit 0;
 }
 
 #Give up root
-&setup_ownership;
+setup_ownership();
 
 # Now set some things from the config file
-# $logfile used to be set in the config file: now we derive it from $logdir
-my $logfile = "$$cfg{logdir}/access.log";
+# $logfile used to be set in the config file: now we derive it from $log_dir
+my $logfile = "$cfg->{log_dir}/access.log";
 
 
 ###################################################
@@ -44,11 +64,12 @@ my $logfile = "$$cfg{logdir}/access.log";
 
 # First we look for rolled and compressed logfiles, from
 # /var/log/apt-cacher/access.log.12.gz to access.log.2.gz
-$logcount = 12;
+my $logcount = 12;
+my @logdata;
 while ($logcount > 1)
 {
 	if (-f "${logfile}.$logcount.gz") {
-		$logdataraw = `zcat ${logfile}.$logcount.gz`;
+		my $logdataraw = `zcat ${logfile}.$logcount.gz`;
 		push (@logdata, split("\n", $logdataraw));
 	}
 	$logcount--;
@@ -56,17 +77,16 @@ while ($logcount > 1)
 
 # Then the immediately rolled (but uncompressed) log
 if (-f "${logfile}.1") {
-	open(LOGFILE, "<${logfile}.1");
-	#@logdata = <LOGFILE>;
-	push(@logdata, <LOGFILE>);
-	close(LOGFILE);
+	open(my $lfh, '<', "${logfile}.1");
+	push(@logdata, <$lfh>);
+	close($lfh);
 }
 
 # Then finally the current working log
 if (-f "${logfile}") {
-	open(LOGFILE, "<$logfile");
-	push(@logdata, <LOGFILE>);
-	close(LOGFILE);
+	open(my $lfh, '<', $logfile);
+	push(@logdata, <$lfh>);
+	close($lfh);
 }
 
 #read current time
@@ -80,24 +100,30 @@ my $hit_count = 0;
 my $hit_bytes = 0;
 my $miss_count = 0;
 my $miss_bytes = 0;
+my ($firstrecord,$lastrecord);
 
 #parse logfile:
-foreach $logfile_line (@logdata)
+foreach (@logdata)
 {
-	#$logfile_line =~ s/ /\+/g;
-	@line = split /\|/, $logfile_line;
-	$req_date = $line[0];
-#	$req_ip   = $line[1];
-	$req_result = $line[2];
-	$req_bytes  = 0;
-	$req_bytes  = $line[3] if $line[3] =~ /^[0-9]+$/;
-#	$req_object = $line[4];
+	my @line = split /\|/;
+	if ($#line==4) { # Old format without PID
+	    splice @line, 1, 0, ''; # Insert empty field
+	}
+
+	next unless $#line == 5; # Skip if parsing fails
+
+	my $req_date = $line[0];
+#	my $req_pid = $line[1];
+#	my $req_ip   = $line[2];
+	my $req_result = $line[3];
+	my $req_bytes  = $line[4] =~ /^\d+$/ ? $line[4] : 0;
+#	my $req_object = $line[5];
 
 	$lastrecord = $req_date;
 	if(!$firstrecord) {
 		$firstrecord = $req_date;
 	}
-	if ( $req_result eq "HIT" )
+	if ( $req_result =~ /HIT|NOTMOD|HEAD/ )
 	{
 		$hit_count++;
 		$hit_bytes += $req_bytes;
@@ -111,6 +137,7 @@ foreach $logfile_line (@logdata)
 }
 
 my $total_count = $hit_count + $miss_count;
+my ($hit_count_percent,$miss_count_percent);
 
 if($total_count eq 0)
 {
@@ -121,7 +148,7 @@ if($total_count eq 0)
 	$miss_count_percent = (int(($miss_count / $total_count) * 10000)) / 100;
 }
 
-$total_bytes = $hit_bytes + $miss_bytes;
+my $total_bytes = $hit_bytes + $miss_bytes;
 
 ##################################################
 # At this point we have hit/miss/total counts, and hit/miss/total traffic
@@ -130,6 +157,8 @@ $total_bytes = $hit_bytes + $miss_bytes;
 # displays as GB.
 # Yes, I know this really should be a subroutine. Sigh. One day. Maybe.
 
+my ($tx,$total_trafficstring,$hit_trafficstring,$miss_trafficstring,$hit_data_percent,$miss_data_percent);
+
 if($total_bytes > 2097152000)
 {
 	$tx = (int(($total_bytes/1073741824) * 1000)) / 1000;
@@ -183,7 +212,7 @@ if(!$lastrecord)
 
 ##################################################
 # spit out the report
-$output = "
+my $output = "
 <html>
 <title>Apt-cacher traffic report</title><style type=\"text/css\"><!--
 a { text-decoration: none; }
@@ -210,7 +239,7 @@ $output .= "<h2 align=\"center\">summary</h2>
 <table border=0 cellpadding=3 cellspacing=1 bgcolor=\"#000000\" align=\"center\" width=\"600\">
 <tr bgcolor=\"#9999cc\"><th bgcolor=\"#9999cc\"> Item </th><th> Value </th></tr>
 <tr bgcolor=\"#cccccc\"><td bgcolor=\"#ccccff\"> Report generated </td><td> $datetime </td></tr>
-<tr bgcolor=\"#cccccc\"><td bgcolor=\"#ccccff\"> Administrator </td><td> <a href=\"mailto:$$cfg{admin_email}\">$$cfg{admin_email}</a> </td></tr>";
+<tr bgcolor=\"#cccccc\"><td bgcolor=\"#ccccff\"> Administrator </td><td> <a href=\"mailto:$cfg->{admin_email}\">$cfg->{admin_email}</a> </td></tr>";
 $output .= "<tr bgcolor=\"#cccccc\"><td bgcolor=\"#ccccff\"> First request </td><td> $firstrecord </td></tr>";
 $output .= "<tr bgcolor=\"#cccccc\"><td bgcolor=\"#ccccff\"> Last request </td><td> $lastrecord </td></tr>";
 $output .= "<tr bgcolor=\"#cccccc\"><td bgcolor=\"#ccccff\"> Total requests </td><td> $total_count </td></tr>";
@@ -227,11 +256,11 @@ $output .= "<h2 align=\"center\">cache efficiency</h2>
 $output .= "</body></html>\n";
 
 #print $output;
-my $report_file = "$$cfg{logdir}/report.html";
+my $report_file = "$cfg->{log_dir}/report.html";
 unlink $report_file;
-open(REPORT,">$report_file") or die "Unable to open $report_file";
-print REPORT "$output\n";
-close REPORT;
+open(my $rfh, '>', $report_file) or die "Unable to open $report_file";
+print $rfh "$output\n";
+close $rfh;
 
 
 exit 0;
